
@file:Suppress("unused")

package ru.dbotthepony.mc.otm.datagen.loot

import com.mojang.datafixers.util.Pair
import it.unimi.dsi.fastutil.objects.Reference2ObjectArrayMap
import it.unimi.dsi.fastutil.objects.Reference2ObjectFunction
import net.minecraft.advancements.critereon.StatePropertiesPredicate
import net.minecraft.data.DataGenerator
import net.minecraft.data.loot.LootTableProvider
import net.minecraft.resources.ResourceLocation
import net.minecraft.world.level.ItemLike
import net.minecraft.world.level.block.Block
import net.minecraft.world.level.block.SlabBlock
import net.minecraft.world.level.block.state.properties.SlabType
import net.minecraft.world.level.storage.loot.LootPool
import net.minecraft.world.level.storage.loot.LootTable
import net.minecraft.world.level.storage.loot.ValidationContext
import net.minecraft.world.level.storage.loot.entries.LootItem
import net.minecraft.world.level.storage.loot.entries.LootPoolSingletonContainer
import net.minecraft.world.level.storage.loot.functions.CopyNbtFunction
import net.minecraft.world.level.storage.loot.functions.SetItemCountFunction
import net.minecraft.world.level.storage.loot.parameters.LootContextParamSet
import net.minecraft.world.level.storage.loot.parameters.LootContextParamSets
import net.minecraft.world.level.storage.loot.predicates.LootItemBlockStatePropertyCondition
import net.minecraft.world.level.storage.loot.providers.nbt.ContextNbtProvider
import net.minecraft.world.level.storage.loot.providers.number.ConstantValue
import ru.dbotthepony.mc.otm.block.entity.MatteryBlockEntity
import ru.dbotthepony.mc.otm.block.entity.MatteryWorkerBlockEntity
import java.util.function.BiConsumer
import java.util.function.Consumer
import java.util.function.Supplier

private typealias LootTableSaver = BiConsumer<ResourceLocation, LootTable.Builder>
private typealias LootTableCallback = Consumer<LootTableSaver>
private typealias LootTableCallbackProvider = Supplier<LootTableCallback>
private typealias LootTuple = Pair<LootTableCallbackProvider, LootContextParamSet>

data class NbtCopy(val source: String, val destination: String, val strategy: CopyNbtFunction.MergeStrategy = CopyNbtFunction.MergeStrategy.REPLACE)

fun TileNbtCopy(source: String, strategy: CopyNbtFunction.MergeStrategy = CopyNbtFunction.MergeStrategy.REPLACE): NbtCopy {
	return NbtCopy(source, "BlockEntityTag.$source", strategy)
}

private val basicTags = arrayOf(
	TileNbtCopy(MatteryBlockEntity.REDSTONE_SIGNAL_KEY),
	TileNbtCopy(MatteryBlockEntity.REDSTONE_SETTING_KEY),
)

private val poweredTags = arrayOf(
	*basicTags,
	TileNbtCopy(MatteryBlockEntity.ENERGY_KEY),
	TileNbtCopy(MatteryBlockEntity.BATTERY_KEY),
)

private val workerTags = arrayOf(
	*poweredTags,
	TileNbtCopy(MatteryWorkerBlockEntity.JOB_KEY),
	TileNbtCopy(MatteryWorkerBlockEntity.WORK_TICKS_KEY),
)

private val poweredMatterWorker = arrayOf(
	*workerTags,
	TileNbtCopy(MatteryBlockEntity.MATTER_STORAGE_KEY),
)

class LootTables(generator: DataGenerator) : LootTableProvider(generator) {
	private val providersTable = Reference2ObjectArrayMap<LootContextParamSet, HashMap<ResourceLocation, () -> LootTable.Builder>>()

	fun builder(context: LootContextParamSet, id: ResourceLocation, provider: LootTable.Builder.() -> Unit) {
		provider(context, id) {
			LootTable.lootTable().also(provider)
		}
	}

	fun provider(context: LootContextParamSet, id: ResourceLocation, provider: () -> LootTable.Builder) {
		check(providersTable
			.computeIfAbsent(context, Reference2ObjectFunction { HashMap() })
			.put(id, provider) == null) { "Duplicate loot pool entry for $id" }
	}

	override fun getTables(): List<LootTuple> {
		return providersTable.entries.stream().map { entry ->
			Pair.of(LootTableCallbackProvider {
				LootTableCallback {
					for ((id, callback) in entry.value) {
						it.accept(id, callback.invoke())
					}
				}
			}, entry.key)
		}.toList()
	}

	override fun validate(map: MutableMap<ResourceLocation, LootTable>, validationtracker: ValidationContext) {}

	fun createSlabItemTable(block: Block, configurator: LootPoolSingletonContainer.Builder<*>.() -> Unit = {}) {
		builder(LootContextParamSets.BLOCK, block.lootTable) {
			withPool(
				LootPool.lootPool().setRolls(ConstantValue.exactly(1.0f)).add(
					LootItem.lootTableItem(block).apply(
						SetItemCountFunction.setCount(
							ConstantValue.exactly(2.0f)
						).`when`(
							LootItemBlockStatePropertyCondition.hasBlockStateProperties(block).setProperties(
								StatePropertiesPredicate.Builder.properties()
									.hasProperty(SlabBlock.TYPE, SlabType.DOUBLE)
							)
						)
					).also(configurator)
				)
			)
		}
	}

	fun createSlabItemTable(blocks: Collection<Block>, configurator: LootPoolSingletonContainer.Builder<*>.() -> Unit = {}) {
		for (block in blocks) {
			createSlabItemTable(block, configurator)
		}
	}

	fun blockProvider(block: Block, provider: () -> LootTable.Builder) {
		provider(LootContextParamSets.BLOCK, block.lootTable, provider)
	}

	fun block(block: Block, provider: LootPool.Builder.() -> Unit) {
		builder(LootContextParamSets.BLOCK, block.lootTable) {
			lootPool(provider)
		}
	}

	fun singleLootPool(context: LootContextParamSet, id: ResourceLocation, block: LootPool.Builder.() -> Unit) {
		builder(context, id) {
			withPool(LootPool.lootPool().also(block))
		}
	}

	fun dropsOther(block: Block, other: ItemLike) {
		singleLootPool(LootContextParamSets.BLOCK, block.lootTable) {
			add(LootItem.lootTableItem(other))
		}
	}

	fun dropsSelf(vararg blocks: Block) {
		for (block in blocks) {
			singleLootPool(LootContextParamSets.BLOCK, block.lootTable) {
				item(block) {}
			}
		}
	}

	fun dropsSelf(block: Block, configurator: LootPoolSingletonContainer.Builder<*>.() -> Unit = {}) {
		singleLootPool(LootContextParamSets.BLOCK, block.lootTable) {
			item(block, configurator)
		}
	}

	fun dropsSelf(blocks: Collection<Block>, configurator: LootPoolSingletonContainer.Builder<*>.() -> Unit = {}) {
		for (block in blocks) {
			singleLootPool(LootContextParamSets.BLOCK, block.lootTable) {
				item(block, configurator)
			}
		}
	}

	fun tile(block: Block, f: (CopyNbtFunction.Builder) -> Unit = {}) {
		singleLootPool(LootContextParamSets.BLOCK, block.lootTable) {
			add(LootItem.lootTableItem(block).also {
				it.apply(CopyNbtFunction.copyData(ContextNbtProvider.BLOCK_ENTITY).also {
					it.copy("Name", "BlockEntityTag.Name")
					f(it)
				})
			})
		}
	}

	fun tile(block: Block, vararg tags: NbtCopy) {
		tile(block) {
			for ((source, destination, strategy) in tags) {
				it.copy(source, destination, strategy)
			}
		}
	}

	fun basicTile(block: Block, vararg tags: NbtCopy) {
		tile(block) {
			for ((source, destination, strategy) in tags) {
				it.copy(source, destination, strategy)
			}

			for ((source, destination, strategy) in basicTags) {
				it.copy(source, destination, strategy)
			}
		}
	}

	fun poweredTile(block: Block, vararg tags: NbtCopy) {
		tile(block) {
			for ((source, destination, strategy) in tags) {
				it.copy(source, destination, strategy)
			}

			for ((source, destination, strategy) in poweredTags) {
				it.copy(source, destination, strategy)
			}
		}
	}

	fun workerTile(block: Block, vararg tags: NbtCopy) {
		tile(block) {
			for ((source, destination, strategy) in tags) {
				it.copy(source, destination, strategy)
			}

			for ((source, destination, strategy) in workerTags) {
				it.copy(source, destination, strategy)
			}
		}
	}

	fun matterWorkerTile(block: Block, vararg tags: NbtCopy) {
		tile(block) {
			for ((source, destination, strategy) in tags) {
				it.copy(source, destination, strategy)
			}

			for ((source, destination, strategy) in poweredMatterWorker) {
				it.copy(source, destination, strategy)
			}
		}
	}

	fun tile(block: Block, vararg tags: String) {
		tile(block, *tags.map { NbtCopy(it, "BlockEntityTag.$it", CopyNbtFunction.MergeStrategy.REPLACE) }.toTypedArray())
	}

	// fix overload resolution by adding extra required argument
	fun basicTile(block: Block, f: String, vararg tags: String) {
		basicTile(block, *tags.map { TileNbtCopy(it) }.toMutableList().also { it.add(TileNbtCopy(f)) }.toTypedArray())
	}

	fun poweredTile(block: Block, f: String, vararg tags: String) {
		poweredTile(block, *tags.map { TileNbtCopy(it) }.toMutableList().also { it.add(TileNbtCopy(f)) }.toTypedArray())
	}

	fun workerTile(block: Block, f: String, vararg tags: String) {
		workerTile(block, *tags.map { TileNbtCopy(it) }.toMutableList().also { it.add(TileNbtCopy(f)) }.toTypedArray())
	}

	fun matterWorkerTile(block: Block, f: String, vararg tags: String) {
		matterWorkerTile(block, *tags.map { TileNbtCopy(it) }.toMutableList().also { it.add(TileNbtCopy(f)) }.toTypedArray())
	}
}
