package ru.dbotthepony.mc.otm.datagen.models

import com.google.gson.JsonArray
import com.google.gson.JsonObject
import com.google.gson.JsonPrimitive
import com.mojang.math.Vector3f
import net.minecraft.client.renderer.block.model.BlockModel
import net.minecraft.core.Direction
import net.minecraft.resources.ResourceLocation
import net.minecraftforge.client.model.generators.ModelBuilder
import net.minecraftforge.common.data.ExistingFileHelper
import ru.dbotthepony.mc.otm.container.set
import ru.dbotthepony.mc.otm.core.set

data class TextureSize(val width: Float, val height: Float) {
	constructor(arr: JsonArray) : this(arr[0].asFloat, arr[1].asFloat)

	init {
		require(width > 0f) { "Invalid width $width" }
		require(height > 0f) { "Invalid width $height" }
	}
}

class MatteryModelBuilder(resourceLocation: ResourceLocation, existingFileHelper: ExistingFileHelper) : ModelBuilder<MatteryModelBuilder>(resourceLocation, existingFileHelper) {
	var textureSize: TextureSize? = null

	fun fromJson(input: JsonObject) {
		input["parent"]?.let {
			parent(ExistingModelFile(ResourceLocation(it.asString), existingFileHelper))
		}

		input["ambientocclusion"]?.let {
			ambientOcclusion = it.asBoolean
		}

		input["gui_light"]?.let {
			guiLight = BlockModel.GuiLight.valueOf(it.asString)
		}

		(input["texture_size"] as? JsonArray)?.let {
			textureSize = TextureSize(it)
		}

		(input["textures"] as? JsonObject)?.let {
			for ((k, v) in it.entrySet()) {
				texture(k, v.asString)
			}
		}

		(input["elements"] as? JsonArray)?.let {
			var i = -1

			for (value in it) {
				i++
				check(value is JsonObject) { "Encountered invalid element at $i" }
				element().fromJson(value)
			}
		}
	}

	override fun toJson(): JsonObject {
		return super.toJson().also {
			val textureSize = textureSize

			if (textureSize != null) {
				it["texture_size"] = JsonArray().also {
					it.add(textureSize.width)
					it.add(textureSize.height)
				}
			}
		}
	}

	override fun element(): MatteryModelElement {
		check(customLoader == null) { "Can not use custom loaders and elements at the same time" }
		return MatteryModelElement().also(elements::add)
	}

	override fun element(index: Int): MatteryModelElement {
		return super.element(index) as MatteryModelElement
	}

	fun offset(x: Float, y: Float, z: Float) {
		for (element in elements) {
			(element as MatteryModelElement).offset(x, y, z)
		}
	}

	fun offset(value: Vector3f) {
		for (element in elements) {
			(element as MatteryModelElement).offset(value)
		}
	}

	inner class MatteryModelElement : ElementBuilder() {
		private var from: Vector3f = Vector3f()
		private var to: Vector3f = Vector3f(16f, 16f, 16f)

		fun from(value: Vector3f) = from(value.x(), value.y(), value.z())
		fun to(value: Vector3f) = to(value.x(), value.y(), value.z())

		override fun from(x: Float, y: Float, z: Float): MatteryModelElement {
			from = Vector3f(x, y, z)
			super.from(x, y, z)
			return this
		}

		override fun to(x: Float, y: Float, z: Float): MatteryModelElement {
			to = Vector3f(x, y, z)
			super.to(x, y, z)
			return this
		}

		fun offset(x: Float, y: Float, z: Float): MatteryModelElement {
			from(x + from.x(), y + from.y(), z + from.z())
			return to(x + to.x(), y + to.y(), z + to.z())
		}

		fun offset(value: Vector3f) = offset(value.x(), value.y(), value.z())

		fun fromJson(input: JsonObject) {
			val from = input["from"] as JsonArray
			val to = input["to"] as JsonArray

			from(from[0].asFloat, from[1].asFloat, from[2].asFloat)
			to(to[0].asFloat, to[1].asFloat, to[2].asFloat)

			(input["faces"] as? JsonObject)?.let {
				for ((k, v) in it.entrySet()) {
					with(face(Direction.valueOf(k.uppercase()))) {
						check(v is JsonObject) { "Element has invalid face at $k" }

						(v["uv"] as? JsonArray)?.also { uv ->
							check(uv.size() == 4) { "Element at $k has invalid number of uvs ${uv.size()}" }
							uvs(uv[0].asFloat, uv[1].asFloat, uv[2].asFloat, uv[3].asFloat)
						}

						(v["rotation"] as? JsonPrimitive)?.also { rotation ->
							when (rotation.asInt) {
								90 -> rotation(FaceRotation.CLOCKWISE_90)
								-90, 270 -> rotation(FaceRotation.COUNTERCLOCKWISE_90)
								180, -180 -> rotation(FaceRotation.UPSIDE_DOWN)
								else -> rotation(FaceRotation.ZERO)
							}
						}

						(v["emissivity"] as? JsonPrimitive)?.asInt?.also { emissivity ->
							emissivity(emissivity)
						}

						(v["ao"] as? JsonPrimitive)?.asBoolean?.also { ao ->
							ao(ao)
						}

						(v["texture"] as? JsonPrimitive)?.also {
							texture(it.asString)
						}
					}
				}
			}
		}
	}
}
