package ru.dbotthepony.mc.otm.capability

import net.minecraft.nbt.CompoundTag
import net.minecraft.server.level.ServerPlayer
import net.minecraft.world.entity.player.Player
import net.minecraft.world.item.ItemStack
import net.minecraftforge.common.capabilities.ForgeCapabilities
import net.minecraftforge.common.util.INBTSerializable
import ru.dbotthepony.mc.otm.core.Decimal
import ru.dbotthepony.mc.otm.core.getDecimal
import ru.dbotthepony.mc.otm.core.getItemStack
import ru.dbotthepony.mc.otm.core.ifPresentK
import ru.dbotthepony.mc.otm.core.set
import ru.dbotthepony.mc.otm.network.FieldSynchronizer
import ru.dbotthepony.mc.otm.registry.StatNames
import ru.dbotthepony.mc.otm.triggers.AndroidBatteryTrigger

class AndroidPowerSource(
	private val ply: Player,
	synchronizer: FieldSynchronizer,
	initialCharge: Decimal,
	maxCharge: Decimal
) : IMatteryEnergyStorage, INBTSerializable<CompoundTag> {
	private var battery by synchronizer.fraction(initialCharge, name = "android battery")
	private var maxBattery by synchronizer.fraction(maxCharge, name = "android max battery")

	var item by synchronizer.item(setter = setter@{ value, access, setByRemote ->
		access.write(value)

		if (ply is ServerPlayer) {
			AndroidBatteryTrigger.trigger(ply, value)
		}
	}, name = "android battery item")

	override fun serializeNBT(): CompoundTag {
		return CompoundTag().also {
			it["battery"] = battery.serializeNBT()
			it["maxBattery"] = maxBattery.serializeNBT()
			it["item"] = item.serializeNBT()
		}
	}

	override fun deserializeNBT(tag: CompoundTag) {
		battery = tag.getDecimal("battery")
		maxBattery = tag.getDecimal("maxBattery")
		item = tag.getItemStack("item")
	}

	fun tick() {
		if (!item.isEmpty && battery < maxBattery) {
			item.getCapability(ForgeCapabilities.ENERGY).ifPresentK {
				if (it is IMatteryEnergyStorage) {
					battery += it.extractEnergyInner(maxBattery - battery, false)
				} else {
					battery += it.extractEnergy(maxBattery - battery, false)
				}
			}
		}
	}

	override fun receiveEnergyInner(howMuch: Decimal, simulate: Boolean): Decimal {
		return receiveEnergyOuter(howMuch, simulate)
	}

	override fun extractEnergyOuter(howMuch: Decimal, simulate: Boolean): Decimal {
		return Decimal.ZERO
	}

	override fun extractEnergyInner(howMuch: Decimal, simulate: Boolean): Decimal {
		@Suppress("name_shadowing")
		var howMuch = howMuch
		var drained = Decimal.ZERO

		if (!item.isEmpty) {
			item.getCapability(ForgeCapabilities.ENERGY).ifPresentK {
				if (it is IMatteryEnergyStorage) {
					val extracted = it.extractEnergyOuter(howMuch, simulate)
					drained += extracted
					howMuch -= extracted
				} else {
					val extracted = it.extractEnergy(howMuch, simulate)
					drained += extracted
					howMuch -= extracted
				}
			}

			if (howMuch.isZero) {
				if (!simulate && ply is ServerPlayer) {
					ply.awardStat(StatNames.POWER_CONSUMED, drained.toInt() * 10)
				}

				return drained
			}
		}

		val new = (battery - howMuch).moreThanZero()
		drained += battery - new

		if (!simulate) {
			battery = new

			if (ply is ServerPlayer) {
				ply.awardStat(StatNames.POWER_CONSUMED, drained.toInt() * 10)
			}
		}

		return drained
	}

	override fun receiveEnergyOuter(howMuch: Decimal, simulate: Boolean): Decimal {
		@Suppress("name_shadowing")
		var howMuch = howMuch
		var received = Decimal.ZERO

		if (!item.isEmpty) {
			item.getCapability(ForgeCapabilities.ENERGY).ifPresentK {
				if (it is IMatteryEnergyStorage) {
					val extracted = it.receiveEnergyOuter(howMuch, simulate)
					received += extracted
					howMuch -= extracted
				} else {
					val extracted = it.receiveEnergy(howMuch, simulate)
					received += extracted
					howMuch -= extracted
				}
			}

			if (howMuch.isZero) {
				return received
			}
		}

		val new = (battery + howMuch).coerceAtMost(maxBattery)
		received += new - battery

		if (!simulate) {
			battery = new
		}

		return received
	}

	override var batteryLevel: Decimal
		get() {
			if (!item.isEmpty) {
				item.getCapability(ForgeCapabilities.ENERGY).ifPresentK {
					if (it is IMatteryEnergyStorage) {
						return battery + it.batteryLevel
					} else {
						return battery + it.energyStored
					}
				}
			}

			return battery
		}
		set(value) {
			battery = value
		}

	override var maxBatteryLevel: Decimal
		get() {
			if (item != ItemStack.EMPTY) {
				item.getCapability(ForgeCapabilities.ENERGY).ifPresentK {
					if (it is IMatteryEnergyStorage) {
						return maxBattery + it.maxBatteryLevel
					} else {
						return maxBattery + it.maxEnergyStored
					}
				}
			}

			return maxBattery
		}
		set(value) {
			maxBattery = value
		}
}
