package ru.dbotthepony.mc.otm.client.screen.panels.buttons

import com.mojang.blaze3d.platform.InputConstants
import com.mojang.blaze3d.vertex.PoseStack
import net.minecraft.ChatFormatting
import net.minecraft.client.gui.screens.Screen
import net.minecraft.network.chat.Component
import ru.dbotthepony.mc.otm.client.render.AbstractSkinElement
import ru.dbotthepony.mc.otm.client.render.UVWindingOrder
import ru.dbotthepony.mc.otm.client.screen.panels.EditablePanel
import ru.dbotthepony.mc.otm.core.GetterSetter
import ru.dbotthepony.mc.otm.core.TextComponent
import ru.dbotthepony.mc.otm.core.next
import ru.dbotthepony.mc.otm.core.prev
import ru.dbotthepony.mc.otm.core.value
import java.util.*
import kotlin.collections.ArrayList

abstract class EnumRectangleButtonPanel<out S : Screen, T : Enum<T>>(
	screen: S,
	parent: EditablePanel<*>?,
	x: Float = 0f,
	y: Float = 0f,
	width: Float,
	height: Float,
	val enum: Class<T>,
	val prop: GetterSetter<T>,
	val defaultValue: T,
	val onChange: ((newValue: T) -> Unit)? = null,
) : RectangleButtonPanel<S>(screen, parent, x, y, width, height, null) {
	private var building = true

	protected val enumMapping = EnumMap<T, Pair<AbstractSkinElement, UVWindingOrder>>(enum)
	protected val tooltipMapping = EnumMap<T, Component>(enum)

	fun addTooltip(value: T, component: Component): EnumRectangleButtonPanel<S, T> {
		check(tooltipMapping.put(value, component) == null) { "Already has mapping for $value" }
		return this
	}

	var mainTooltip: Component? = null
		set(value) {
			if (field != null) {
				throw UnsupportedOperationException("Write once only")
			}

			field = value
		}

	fun isFullyDefined(): Boolean {
		for (value in enum.enumConstants) {
			if (enumMapping[value] == null) {
				return false
			}
		}

		return true
	}

	val missingValues: List<T> get() {
		val missing = ArrayList<T>()

		for (value in enum.enumConstants) {
			if (enumMapping[value] == null) {
				missing.add(value)
			}
		}

		return missing
	}

	fun add(value: T, skinElement: AbstractSkinElement, winding: UVWindingOrder = UVWindingOrder.NORMAL): EnumRectangleButtonPanel<S, T> {
		return add(value, skinElement to winding)
	}

	fun add(value: T, skinElement: AbstractSkinElement, component: Component, winding: UVWindingOrder = UVWindingOrder.NORMAL): EnumRectangleButtonPanel<S, T> {
		return add(value, component, skinElement to winding)
	}

	fun add(value: T, component: Component, skinElement: AbstractSkinElement, winding: UVWindingOrder = UVWindingOrder.NORMAL): EnumRectangleButtonPanel<S, T> {
		return add(value, component, skinElement to winding)
	}

	fun add(value: T, pair: Pair<AbstractSkinElement, UVWindingOrder>): EnumRectangleButtonPanel<S, T> {
		check(building) { "Not building" }
		check(enumMapping.put(value, pair) == null) { "Already has mapping for $value" }

		if (enumMapping.size == enum.enumConstants.size) {
			finish()
		}

		return this
	}

	fun add(value: T, component: Component, pair: Pair<AbstractSkinElement, UVWindingOrder>): EnumRectangleButtonPanel<S, T> {
		addTooltip(value, component)
		return add(value, pair)
	}

	fun finish(): EnumRectangleButtonPanel<S, T> {
		check(building) { "Not building" }
		check(isFullyDefined()) {
			"Not all enums having their mapping defined, missing are: ${missingValues.joinToString(", ")}"
		}

		building = false
		return this
	}

	override fun innerRender(stack: PoseStack, mouseX: Float, mouseY: Float, partialTick: Float) {
		check(!building) { "Still building this button!" }
		super.innerRender(stack, mouseX, mouseY, partialTick)
		val pair = checkNotNull(enumMapping[prop.get()]) { "HOW" }
		pair.first.render(stack, 0f, 0f, width, height, pair.second)
	}

	override fun mouseClickedInner(x: Double, y: Double, button: Int): Boolean {
		check(!building) { "Still building this button!" }
		return super.mouseClickedInner(x, y, button)
	}

	override fun mouseReleasedInner(x: Double, y: Double, button: Int): Boolean {
		check(!building) { "Still building this button!" }
		return super.mouseReleasedInner(x, y, button)
	}

	override fun onClick(clickButton: Int) {
		when (clickButton) {
			InputConstants.MOUSE_BUTTON_LEFT -> {
				prop.value = prop.value.next(enum.enumConstants)
				onChange?.invoke(prop.get())
			}

			InputConstants.MOUSE_BUTTON_RIGHT -> {
				prop.value = prop.value.prev(enum.enumConstants)
				onChange?.invoke(prop.get())
			}

			InputConstants.MOUSE_BUTTON_MIDDLE -> {
				if (prop.get() != defaultValue) {
					prop.value = defaultValue
					onChange?.invoke(prop.get())
				}
			}
		}
	}

	override fun innerRenderTooltips(stack: PoseStack, mouseX: Float, mouseY: Float, partialTick: Float): Boolean {
		if (!isHovered && !isGrabbingMouseInput()) {
			return super.innerRenderTooltips(stack, mouseX, mouseY, partialTick)
		}

		if (mainTooltip == null && tooltipMapping.size == 0) {
			return super.innerRenderTooltips(stack, mouseX, mouseY, partialTick)
		}

		val listing = ArrayList<Component>()

		if (mainTooltip != null) {
			listing.add(mainTooltip!!)
			listing.add(SPACE)
		}

		for ((key, value) in tooltipMapping) {
			if (key == prop.get()) {
				listing.add(value.copy().withStyle(ChatFormatting.WHITE))
			} else {
				listing.add(value.copy().withStyle(ChatFormatting.GRAY))
			}
		}

		screen.renderComponentTooltip(
			stack,
			listing,
			mouseX.toInt(),
			mouseY.toInt(),
			font
		)

		return true
	}

	companion object {
		private val SPACE = TextComponent("")
	}
}
